﻿// -------------------------------------------------------------------------------------------------
// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
// -------------------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.IO;
using Hl7.Fhir.Model;
using Hl7.Fhir.Serialization;
using Microsoft.Health.Core.Internal;
using Microsoft.Health.Fhir.Core.Extensions;
using Microsoft.Health.Fhir.Core.Features.Context;
using Microsoft.Health.Fhir.Core.Features.Persistence;
using Microsoft.Health.Fhir.Core.Features.Routing;
using Microsoft.Health.Fhir.Core.Features.Search;
using Microsoft.Health.Fhir.Core.Models;
using Microsoft.Health.Fhir.Shared.Core.Features.Search;
using Microsoft.Health.Fhir.Tests.Common;
using Microsoft.Health.Test.Utilities;
using NSubstitute;
using Xunit;

using Task = System.Threading.Tasks.Task;

namespace Microsoft.Health.Fhir.Core.UnitTests.Features.Search
{
    public class BundleFactoryTests
    {
        private readonly FhirJsonSerializer _fhirJsonSerializer = new FhirJsonSerializer();
        private readonly IUrlResolver _urlResolver = Substitute.For<IUrlResolver>();
        private readonly IFhirRequestContextAccessor _fhirRequestContextAccessor = Substitute.For<IFhirRequestContextAccessor>();
        private readonly BundleFactory _bundleFactory;

        private const string _continuationToken = "ct";
        private const string _resourceUrlFormat = "http://resource/{0}";
        private static readonly string _correlationId = Guid.NewGuid().ToString();
        private static readonly Uri _selfUrl = new Uri("http://self/");
        private static readonly Uri _nextUrl = new Uri("http://next/");
        private static readonly IReadOnlyList<Tuple<string, string>> _unsupportedSearchParameters = new Tuple<string, string>[0];
        private static readonly IReadOnlyList<(string searchParameter, string reason)> _unsupportedSortingParameters = Array.Empty<(string parameterName, string reason)>();

        private static readonly DateTimeOffset _dateTime = new DateTimeOffset(2019, 1, 5, 15, 30, 23, TimeSpan.FromHours(8));

        public BundleFactoryTests()
        {
            _bundleFactory = new BundleFactory(
                _urlResolver,
                _fhirRequestContextAccessor);

            IFhirRequestContext fhirRequestContext = Substitute.For<IFhirRequestContext>();

            fhirRequestContext.CorrelationId.Returns(_correlationId);

            _fhirRequestContextAccessor.FhirRequestContext.Returns(fhirRequestContext);
        }

        [Fact]
        public void GivenAnEmptySearchResult_WhenCreateSearchBundle_ThenCorrectBundleShouldBeReturned()
        {
            _urlResolver.ResolveRouteUrl(_unsupportedSearchParameters, _unsupportedSortingParameters).Returns(_selfUrl);

            ResourceElement actual = null;

            using (Mock.Property(() => ClockResolver.UtcNowFunc, () => _dateTime))
            {
                actual = _bundleFactory.CreateSearchBundle(new SearchResult(new SearchResultEntry[0], _unsupportedSearchParameters, _unsupportedSortingParameters, null));
            }

            Assert.NotNull(actual);
            Assert.Equal(Bundle.BundleType.Searchset.ToString().ToLowerInvariant(), actual.Scalar<string>("Bundle.type"));
            Assert.Equal(_correlationId.ToString(), actual.Id);
            Assert.Equal(_dateTime, actual.LastUpdated);
            Assert.Equal(_selfUrl.OriginalString, actual.Scalar<string>("Bundle.link.where(relation='self').url"));
        }

        [Fact]
        public void GivenASearchResult_WhenCreateSearchBundle_ThenCorrectBundleShouldBeReturned()
        {
            _urlResolver.ResolveResourceWrapperUrl(Arg.Any<ResourceWrapper>()).Returns(x => new Uri(string.Format(_resourceUrlFormat, x.ArgAt<ResourceWrapper>(0).ResourceId)));
            _urlResolver.ResolveRouteUrl(_unsupportedSearchParameters, _unsupportedSortingParameters).Returns(_selfUrl);

            ResourceElement observation1 = Samples.GetDefaultObservation().UpdateId("123");
            ResourceElement observation2 = Samples.GetDefaultObservation().UpdateId("abc");

            var resourceWrappers = new SearchResultEntry[]
            {
                new SearchResultEntry(CreateResourceWrapper(observation1)),
                new SearchResultEntry(CreateResourceWrapper(observation2)),
            };

            var searchResult = new SearchResult(resourceWrappers, _unsupportedSearchParameters, _unsupportedSortingParameters, continuationToken: null);

            ResourceElement actual = null;

            using (Mock.Property(() => ClockResolver.UtcNowFunc, () => _dateTime))
            {
                actual = _bundleFactory.CreateSearchBundle(searchResult);
            }

            // Since there is no continuation token, there should not be next link.
            Assert.Null(actual.Scalar<string>("Bundle.link.where(relation='next').url"));
            Assert.Collection(
                actual.ToPoco<Bundle>().Entry,
                async e => await ValidateEntry(observation1.ToPoco<Observation>(), e),
                async e => await ValidateEntry(observation2.ToPoco<Observation>(), e));

            ResourceWrapper CreateResourceWrapper(ResourceElement resourceElement)
            {
                return new ResourceWrapper(
                    resourceElement,
                    new RawResource(_fhirJsonSerializer.SerializeToString(resourceElement.ToPoco<Observation>()), FhirResourceFormat.Json, isMetaSet: false),
                    null,
                    false,
                    null,
                    null,
                    null);
            }

            async Task ValidateEntry(Observation expected, Bundle.EntryComponent actualEntry)
            {
                Assert.NotNull(actualEntry);

                var raw = actualEntry as RawBundleEntryComponent;

                using (var ms = new MemoryStream())
                using (var sr = new StreamReader(ms))
                {
                    await raw?.ResourceElement?.SerializeToStreamAsUtf8Json(ms);
                    ms.Seek(0, SeekOrigin.Begin);
                    var resourceData = await sr.ReadToEndAsync();
                    Assert.NotNull(resourceData);

                    Resource resource;
                    resource = new FhirJsonParser().Parse(resourceData) as Resource;

                    Assert.Equal(expected.Id, resource.Id);
                    Assert.Equal(string.Format(_resourceUrlFormat, expected.Id), raw.FullUrl);
                    Assert.NotNull(raw.Search);
                    Assert.Equal(Bundle.SearchEntryMode.Match, raw.Search.Mode);
                }
            }
        }

        [Fact]
        public void GivenASearchResultWithContinuationToken_WhenCreateSearchBundle_ThenCorrectBundleShouldBeReturned()
        {
            string encodedContinuationToken = Convert.ToBase64String(System.Text.Encoding.UTF8.GetBytes(_continuationToken));
            _urlResolver.ResolveRouteUrl(_unsupportedSearchParameters, _unsupportedSortingParameters, encodedContinuationToken, true).Returns(_nextUrl);
            _urlResolver.ResolveRouteUrl(_unsupportedSearchParameters, _unsupportedSortingParameters).Returns(_selfUrl);

            var searchResult = new SearchResult(new SearchResultEntry[0], _unsupportedSearchParameters, _unsupportedSortingParameters, _continuationToken);

            ResourceElement actual = _bundleFactory.CreateSearchBundle(searchResult);

            Assert.Equal(_nextUrl.OriginalString, actual.Scalar<string>("Bundle.link.where(relation='next').url"));
        }
    }
}
